% Preprocessing of the ith column of M obtained by solving the problem 
%
% min_x ||M(:,i) - M_I x||_2^2 
%
% such that M(:,i) + epsilon*||M(:,i)||_inf  >= M_I x and x >= 0, 
%
% where M_I = M(:,I), I = {1,2,...,n}\{i}. 
%
% See Nicolas Gillis, "Sparse and Unique Nonnegative Matrix Factorization 
% Through Data Preprocessing", arXiv. 
%
%
% *** Input ***
%   M              : (m x n) data matrix.
%   1 <= i <= n    : index of the column to preprocess
%   epsilon >= 0   : relxation allowed
%   solver         : = 1, lsqlin (default) 
%                    = 2, quadprog active set
%                    = 3, quadprog interior point
%
% *** Output ***
%   
%   x                  : optimal solution of the CLLS problem above
%   m  = M_{:i}-M_I x  : preprocessed ith column of M

function [x,m] = preprocNMFonecol(M,i,epsilon,solver); 

tolfin = 0; 
[m,r] = size(M); 
if nargin <= 2, epsilon = 0; else epsilon = epsilon*max(M(:,i))*ones(m,1); end
if nargin <= 3, solver = 1; end 

ind = [1:i-1 i+1:r]; 
C = M(:,ind); d = M(:,i); 
OPTIONS.MaxIter = 1000; 
OPTIONS.LargeScale = 'off'; 
OPTIONS.Display = 'off'; 

% Different solvers are possible:
if solver == 2 || solver == 3 
    % (2) Quadprog with active-set
    if solver == 2,  solver = 'active-set'; end
    % (3) Quadprog with interior-point; 
    if solver == 3,  solver = 'interior-point-convex'; end
    OPTIONS.Algorithm = solver; 
    x = quadprog(C'*C,-C'*d,C,d+epsilon,[],[],zeros(r-1,1),[],[],OPTIONS); 
else
    % (1) lsqlin 
    x = lsqlin(C,d,C,d+epsilon,[],[],zeros(r-1,1),[],[],OPTIONS); 
end
m = d - C*x; 