% recursive Sparse NMU algorithm
%
% cf. "Sparse Nonnegative Matrix Underapproximation and its Application to Hyperspectral Image Analysis", 
% N. Gillis and R.J. Plemmons, Preprint, 2011.
%
% website. http://sites.google.com/site/nicolasgillis/
%
% [x,y] = recursiveSNMU(M,r,lami,delta,Delta,maxiter,init)
%
% Input.
%   M              : (m x n) matrix to factorize.
%   r              : factorization rank, default = 1.
%   lami           : parameter for sparsity, in [0,1[, default = 0.2.
%   delta          : lower bound for sparsity, in [0,1[, default = 0.01.
%   Delta          : upper bound for sparsity, in [delta,1], default = 1.
%   maxiter        : number of iterations, default = 100.
%   init           : =1: svds based initialization (default), otherwise random
%
% Output.
%   (U,V) : solution, UV^T "<=" M, U >= 0 sparse, V >= 0 ("."-approximately)

function [U,V] = recursiveSNMU(M,r,lami,delta,Delta,maxiter,init)

[m,n] = size(M);
if nargin <= 6, init = 1; end
if nargin <= 5, maxiter = 100; end
if nargin <= 4, Delta = 1; end
if nargin <= 3, delta = 0.01; end
if nargin <= 2, lam0 = 0.2; end
if nargin <= 1, r = 1; end

% NMU Recursion
for k = 1 : r
    if length(lami) == r
        lam0 = lami(k);
    else
        lam0 = lami(1);
    end
    fprintf('Step %d of the sparse NMU recursion started \n',k); 
    % Initialization of (x,y,sigma)
    x = rand(m,1); y = (M'*x); y = y/norm(y);
    x = (M*y); sigma = norm(x); x = x/sigma;
    if init == 1
        % Initialization of (x,y,sigma) with optimal rank-one NMF of M using power method
        for in = 1 : maxiter
            y = (M'*x); y = y/norm(y);
            x = (M*y); sigma = norm(x); x = x/sigma;
        end
    end
    U(:,k) = x; V(:,k) = sigma*y; 
    % Initialization of Lagrangian variable Lambda
    % -Performed by block to decrease memory requirment
    indic = 1:floor(m/2); 
    Lambda(indic,:) = max(0,sigma*x(indic)*y'-M(indic,:)); 
    indic = floor(m/2)+1:m;
    Lambda(indic,:) = max(0,sigma*x(indic)*y'-M(indic,:)); 
    % Alternating optimization
    j = 0; 
    while (j < maxiter) && (j <= 1 || res(j) > 1e-4)
        j = j + 1; 
        xp = x; yp = y; 
        for p = 1 : 1
            % ***(x,y) Update***
           	if j == 1
                x = max(0,(M*y-Lambda*y)); 
                lam = lam0*max(x); 
                x = max(0,(x-lam)); x = x/norm(x); 
            else
                x = max(0,(M*y-Lambda*y)); X = max(x); 
                if X < lam
                    lam = max(0.01,0.99*X); disp('mu changed bcs x equals zero');
                end
                x = max(0,(x-lam)); x = x/norm(x); 
            end
            if sum(x>0) < delta*m
                lam = lam*0.95; disp('mu decreased');
            elseif sum(x>0) > Delta*m
                lam = lam*1.05; disp('mu increased');
            end
            y = max(0,(M'*x-Lambda'*x));  y = y/norm(y); 
            sigma = ((x'*M-x'*Lambda)*y);
            mu(j) = lam/sigma; 
            % Check if the solution is non-trivial
            if sigma > 0
                mX = max(x); U(:,k) = x/mX; 
                V(:,k) = y*mX*sigma; 
            else
                disp('Lambda changed'); 
                Lambda = 0.95*Lambda;  
                y = V(:,k); % Take y as the previous value
            end
        end
        sp(j) = sum(x==0)/m; % Sparsity of x at iteration j
        res(j) = (norm(xp-x) +  norm(yp-y))/2; % Convergence criterion
        if j == 1 
            xs = x; ys = y; sigmas = sigma; jmin = j; 
        elseif res(j) < res(jmin) && ~(1.75*sp(j) < sp(jmin))
            jmin = j; xs = x; ys = y; sigmas = sigma;
        elseif j == maxiter-1 && res(j) > 10*res(jmin)
            fprintf('Algorithm did not converge, best ''stationary point'' kept (%dth iteration) \n', jmin);
            x = xs; y = ys; sigmas = sigma;
            mX = max(x); U(:,k) = x/mX; V(:,k) = sigmas*y*mX; j = maxiter; 
        end
        % ***Lambda Update*** 
        indic = 1:floor(m/2); 
        Lambda(indic,:) = max(0,Lambda(indic,:) + (sigma*x(indic)*y'-M(indic,:))/((j+1))); 
        indic = floor(m/2)+1:m; 
        Lambda(indic,:) = max(0,Lambda(indic,:) + (sigma*x(indic)*y'-M(indic,:))/((j+1)));
        clear A; 
    end
    clear Lambda; 
    M = max(0,M-U(:,k)*V(:,k)');
end